import 'dart:async';
import 'package:async/async.dart';

typedef OnPreExecute = void Function();
typedef OnPostExecute<Result> = void Function({Result result}); 
typedef OnCompleted = void Function();
typedef OnCanceled = void Function();
typedef OnProgressUpdate<Progress> = void Function({List<Progress> values});
typedef DoInBackground<Param, Result> = Future<Result> Function({List<Param> values});

class AsyncTask<Param, Progress, Result> extends CancelableCompleter<Result>{

  OnPreExecute _onPreExecute;
  DoInBackground<Param, Result> _doInBackground;
  OnProgressUpdate<Progress> _onProgressUpdate;
  OnPostExecute<Result> _onPostExecute;
  OnCompleted _onCompleted;
  OnCanceled _onCanceled;

  bool _isCompleted;

  AsyncTask({
    OnPreExecute onPreExecute,
    DoInBackground<Param, Result> doInBackground,
    OnProgressUpdate<Progress> onProgressUpdate,
    OnPostExecute<Result> onPostExecute,
    OnCompleted onCompleted,
    OnCanceled onCanceled,
  }) :
    _isCompleted = false,
    _onPreExecute = onPreExecute,
    _doInBackground = doInBackground,
    _onProgressUpdate = onProgressUpdate,
    _onPostExecute = onPostExecute,
    _onCompleted = onCompleted,
    _onCanceled = onCanceled,
    super(onCancel: onCanceled)
  ;

  @override
  void complete([value]) async{
    super.complete(value);
    _isCompleted = isCanceled ? false : true;
  }

  void publishProgress({List<Progress> values}) => _onProgressUpdate != null ? _onProgressUpdate(values: values) : null;

  void execute({
    OnPreExecute onPreExecute,
    DoInBackground<Param, Result> doInBackground,
    OnProgressUpdate<Progress> onProgressUpdate,
    OnPostExecute<Result> onPostExecute,
    OnCompleted onCompleted,
    List<Param> values,
  }) {
    _onPreExecute = onPreExecute ?? _onPreExecute;
    _doInBackground = doInBackground ?? _doInBackground;
    _onProgressUpdate = onProgressUpdate ?? _onProgressUpdate;
    _onPostExecute = onPostExecute ?? _onPostExecute;
    _onCompleted = onCompleted ?? _onCompleted;
    if(_onPostExecute != null) operation.value.then((value) => _onPostExecute(result: value));
    if(_onCompleted != null) operation.value.whenComplete(_onCompleted);
    if(_onPreExecute != null) _onPreExecute();
    complete(Future(() => _doInBackground(values: values)));
  }

  bool get isRunning => !_isCompleted && !isCanceled;
  @override
  bool get isCompleted => _isCompleted;

  OnPreExecute get onPreExecute => _onPreExecute;
  DoInBackground<Param, Result> get doInBackground => _doInBackground;
  OnProgressUpdate<Progress> get onProgressUpdate => _onProgressUpdate;
  OnPostExecute<Result> get onPostExecute => _onPostExecute;
  OnCompleted get onCompleted => _onCompleted;
  OnCanceled get onCanceled => _onCanceled;

  void cancel() => operation.cancel();
}